/**********************************************************************
**  Copyright 2003 Steve Essery (tecknobabble@hotmail.com)
**
**  This file is part of the TaskList Applet.
**
**  The TaskList Applet is free software; you can redistribute it
**  and/or modify it under the terms of the GNU General Public License
**  as published by the Free Software Foundation; either version 2 of
**  the License, or (at your option) any later version.
**
**  The TaskList Applet is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with the TaskList Applet; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
**********************************************************************/

#include "custommenuwidget.h"
#include "tasklist.h"

#include <qpe/resource.h>
#include <qpainter.h>
#include <qpopupmenu.h>
#include <qapplication.h>
#include <qpe/qpeapplication.h>
#include <qdir.h>
#include <qfile.h>
#include <qpe/config.h>
#include <qevent.h>
#include <qpixmapcache.h>
#include <qlayout.h>
#include <qtimer.h>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <signal.h>


/***
 *
 * logMessage()
 *
 */
static FILE *debugFile = NULL;
void logMessage(const char* format, ...)
{
    if (debugFile != 0)
    {
        va_list ap;
	va_start(ap, format);
	vfprintf(debugFile, format, ap);
	fprintf(debugFile, "\n");
	va_end(ap);
        fflush(debugFile);
    }
}


/***
 *
 * TaskListApplet::TaskListApplet()
 *
 */
TaskListApplet::TaskListApplet( QWidget *parent, const char *name )
    : QWidget( parent, name ),
      channel(NULL),
      menu(NULL), idMap(NULL), channelMap(NULL), taskBarParent(NULL),
      taskBarLayout(NULL), hideTimer(NULL), taskBar(NULL),
      taskBarParentLayout(NULL),
      menuContents("TextAndIcon"), taskBarTimeout(5000),
      showCancel(false), useKill(false), taskBarLeftToRight(false),
      debug(false), canRaiseDesktop(true)
{
    logMessage("ENTER: TaskListApplet");

    // Read in our configuration file information
    readConfig();

    // Get our icon
    logMessage("TaskListApplet: loading tasklist pixmap");
    tasklistPixmap = Resource::loadPixmap( "tasklist/tasklist" );
    setFixedWidth( tasklistPixmap.width() );

    // Maps menu id to qcop channel name
    logMessage("TaskListApplet: about to create idMap");
    idMap = new IDMap();

    // Records which channels we already have menu entries for as an
    // application could be installed on multiple tabs and we only require
    // a single menu entry in these cases
    logMessage("TaskListApplet: about to create channelMap");
    channelMap = new ChannelMap();

    // Create our custom qcop channel
    logMessage("TaskListApplet: about to create qcop channel");
    channel = new QCopChannel("QPE/TaskList", this);
    connect(channel, SIGNAL(received(const QCString&, const QByteArray&)),
            this, SLOT(signalReceived(const QCString&, const QByteArray&)));
    logMessage("EXIT: TaskListApplet");
}


/***
 *
 * TaskListApplet::~TaskListApplet()
 *
 */
TaskListApplet::~TaskListApplet()
{
    logMessage("ENTER: ~TaskListApplet");

    logMessage("~TaskListApplet: about to delete idMap");
    if (idMap) delete idMap;

    logMessage("~TaskListApplet: about to delete channelMap");
    if (channelMap) delete channelMap;

    logMessage("~TaskListApplet: about to delete menu");
    if (menu) delete menu;

    logMessage("~TaskListApplet: about to delete taskBarLayout");
    if (taskBarLayout) delete taskBarLayout;

    logMessage("~TaskListApplet: about to delete taskBar");
    if (taskBar) delete taskBar;

    logMessage("~TaskListApplet: about to delete taskBarParentLayout");
    if (taskBarParentLayout) delete taskBarParentLayout;

    logMessage("~TaskListApplet: about to delete taskBarParent");
    if (taskBarParent) delete taskBarParent;

    //logMessage("~TaskListApplet: about to delete channel");
    //if (channel) delete channel;
    //closing the file here crashes the applet
    //logMessage("~TaskListApplet: about to close debug file");
    //if (debugFile) fclose(debugFile);

    logMessage("EXIT: ~TaskListApplet");
}


/***
 *
 * TaskListApplet::rebuildTaskBar()
 *
 */
void TaskListApplet::rebuildTaskBar()
{
  logMessage("ENTER: rebuildTaskBar()");

  if (taskBarParent)
      if (taskBarParent->isVisible())
          taskBarParent->hide();

  logMessage("rebuildTaskBar: about to delete taskBarLayout");
  if (taskBarLayout) delete taskBarLayout;

  logMessage("rebuildTaskBar: about to delete taskBar");
  if (taskBar) delete taskBar;

  logMessage("rebuildTaskBar: about to delete taskBarParentLayout");
  if (taskBarParentLayout) delete taskBarParentLayout;

  logMessage("rebuildTaskBar: about to delete taskBarParent");
  if (taskBarParent) delete taskBarParent;

  logMessage("rebuildTaskBar: creating parent widget");
  taskBarParent = new QWidget(NULL,
			      NULL,
			      QWidget::WStyle_Customize  |
			      QWidget::WStyle_Tool       |
			      QWidget::WStyle_StaysOnTop |
			      QWidget::WGroupLeader);
  taskBarParent->setGeometry(0, 0, qApp->desktop()->width(), 20);
  taskBarParent->hide();
  taskBarParent->setFocusPolicy(NoFocus);
  taskBarParent->setCaption("TaskListApplet");

  logMessage("rebuildTaskBar: creating parent layout");
  taskBarParentLayout = new QHBoxLayout(taskBarParent, 0, 1);

  // From the behaviour of this it appears someone at TrollTech got their
  // Left and their Right the wrong way round
  if (taskBarLeftToRight)
      taskBarParentLayout->setDirection(QBoxLayout::RightToLeft);
  else
      taskBarParentLayout->setDirection(QBoxLayout::LeftToRight);
  taskBarParentLayout->setAutoAdd(true);

  logMessage("rebuildTaskBar: creating taskbar");
  taskBar = new QWidget(taskBarParent);

  logMessage("rebuildTaskBar: creating taskbar layout");
  taskBarLayout = new QHBoxLayout(taskBar, 0, 1);
  // From the behaviour of this it appears someone at TrollTech got their
  // Left and their Right the wrong way round
  if (taskBarLeftToRight)
      taskBarLayout->setDirection(QBoxLayout::RightToLeft);
  else
      taskBarLayout->setDirection(QBoxLayout::LeftToRight);
  taskBarLayout->setAutoAdd(true);

  logMessage("EXIT: rebuildTaskBar()");
}

/***
 *
 * TaskListApplet::mousePressEvent()
 *
 */
void TaskListApplet::mousePressEvent( QMouseEvent *)
{
    logMessage("ENTER: mousePressEvent");

    if (menuContents == "TaskBar" && taskBarParent)
    {
        if (taskBarParent->isVisible())
	{
	    logMessage("mousePressEvent: TaskBar already visible");
	    logMessage("EXIT: mousePressEvent");
	    return;
	}
    }

    // Call the procedure that does all the work
    rebuildMenu();

    if (menuContents == "TaskBar")
    {
	logMessage("mousePressEvent: about to resize taskbar");
	taskBarParent->resize(qApp->desktop()->width(),
			      taskBarParent->height());
	taskBarParent->move(0, mapToGlobal( QPoint()).y() -
			       taskBarParent->height());
	logMessage("mousePressEvent: about to show taskbar");
	taskBarParent->show();
	repaint();
	if (!hideTimer)
	{
	    logMessage("mousePressEvent: creating hide timer");
            hideTimer = new QTimer(this);
	    connect(hideTimer, SIGNAL(timeout()), SLOT(hideTaskBar()));
	}
	logMessage("mousePressEvent: starting hide timer");
        hideTimer->start(taskBarTimeout, true);
    }
    else
    {
	// Where any items inserted into the menu?
	if (idMap->isEmpty())
	{
	    // No so display something to pad the menu
	    logMessage("mousePressEvent: creating empty menu");
	    int id = menu->insertItem("----");
	    menu->setItemEnabled(id, FALSE);
	}

	// Work out where to display ourselves
	logMessage("mousePressEvent: work out where to popup menu");
	QPoint curPos = mapToGlobal( QPoint(0,0) );
	QSize sh = menu->sizeHint();
	logMessage("mousePressEvent: pop it up");
	menu->popup( curPos-QPoint((sh.width()-width())/2,sh.height()) );
    }

    logMessage("EXIT: mousePressEvent");
}


/***
 *
 * TaskListApplet::rebuildMenu()
 *
 */
void TaskListApplet::rebuildMenu()
{
    QDir    appsDir;
    QFile   dir;
    QString dirPath;

    logMessage("ENTER: rebuildMenu");

    if (menuContents == "TaskBar")
    {
        rebuildTaskBar();
    }
    else
    {
	// Create the popup menu
	if (!menu)
	{
	    // Create a popup menu just the once
	    menu = new QPopupMenu(this);
	}
	else
	{
	    // Clear the menu prior to inserting new items
	    menu->clear();
	}
    }

    // OK...
    // PLAN OF ACTION:
    // 1) Find what applications there could be by scanning
    //    the app files in the dirs under $QPEDIR/apps/*/
    // 2) We need to extract the Icon and Exec information from the
    //    *.desktop files
    // 3) Probe which applications are registered
    //    by using QCopEnvelope::isRegistered(QString channelName)

    // Clear down the channelMap
    logMessage("rebuildMenu: about to clear channelMap");
    if (!channelMap)
        channelMap->clear();
    else
        channelMap = new ChannelMap();

    // Clear down the idMap
    logMessage("rebuildMenu: about to clear idMap");
    if (!idMap)
        idMap->clear();
    else
        idMap = new IDMap();

    // Generate our own item ids
    int menuItemId = 0;

    // Reset the pickedId for cycling through apps
    pickedId = 0;

    // Find the directories under $QPEDIR/apps
    dirPath = QPEApplication::qpeDir();
    dirPath.append("apps");

    // Open the directory looking for the subdirs
    appsDir = QDir(dirPath);
    appsDir.setFilter(QDir::Dirs);

    logMessage("rebuildMenu: about to scan dirs in $QPEDIR/apps");
    // Get an interator for the list of dirs
    const QFileInfoList *appsList = appsDir.entryInfoList();
    QFileInfoListIterator appsIterator( *appsList );
    QFileInfo *appsInfo;

    // Now for each directory...
    while ( (appsInfo = appsIterator.current()) )
    {
        // Move to the next directory for the next loop
        ++appsIterator;

        // Skip . and .. dirs
	if (appsInfo->fileName() == "." || appsInfo->fileName() == "..")
	{
	    logMessage("rebuildMenu: skipping %s", appsInfo->fileName().latin1());
	    continue;
	}
	logMessage("rebuildMenu: about to look for .desktop files in %s", appsInfo->fileName().latin1());

        // Open it looking for the .desktop files
	QString subDirPath = dirPath;
	QString subFilePath;
	subDirPath.append("/");
	subDirPath.append(appsInfo->fileName());
	QDir subDir = QDir(subDirPath);
	subDir.setNameFilter("*.desktop");
	subDir.setFilter(QDir::Files);

        // Again, iterate over the list
	const QFileInfoList *subList = subDir.entryInfoList();
	QFileInfoListIterator subIterator(*subList);
	QFileInfo *subInfo;

        // And for each one...
	while ((subInfo = subIterator.current()))
	{
            QString configName, appName, execName, iconName;
            QString channelName = "QPE/Application/";
	    Config *cfgFile;

	    // Move to the next .desktop file for the next loop
	    ++subIterator;

	    logMessage("rebuildMenu: about to open %s", subInfo->fileName().latin1());

            // Open the config file and look for the information we need
	    configName = subDirPath;
	    configName.append("/");
	    configName.append(subInfo->fileName());
            cfgFile = new Config(configName, Config::File);
            cfgFile->setGroup("Desktop Entry");
            appName  = cfgFile->readEntry("Name");
            execName = cfgFile->readEntry("Exec");
            iconName = cfgFile->readEntry("Icon"); 
            channelName.append(execName);

	    // Check to see if we already have this channel mapped
	    if (!channelMap->find(channelName))
	    {
	        // Is the channel registered?
	        if (QCopChannel::isRegistered((const char *)channelName))
	        {
		    logMessage("rebuildMenu: channel is registered");
		    logMessage("rebuildMenu: appName=%s", appName.latin1());
		    logMessage("rebuildMenu: execName=%s", execName.latin1());
		    logMessage("rebuildMenu: iconName=%s", iconName.latin1());
		    logMessage("rebuildMenu: channelName=%s", channelName.latin1());

		    QPixmap *cachedPixmap = NULL;
		    QPixmap  menuPixmap;

		    // Have we already cached the pixmap? (saves us having to 
		    // resize the icons every time the menu is popped up)
		    if ((cachedPixmap = QPixmapCache::find(iconName)))
		    {
			// Use the cached pixmap
			menuPixmap = *cachedPixmap;
		    }
		    else
		    {
			// Otherwise load as a image and smooth scale it to the 
			// required size.
			QImage im1 = Resource::loadImage(iconName);
			im1 = im1.smoothScale(16,16);

			// Convert the image to a pixmap and cache it
			menuPixmap.convertFromImage(im1);
			cachedPixmap = new QPixmap(menuPixmap);
			QPixmapCache::insert(iconName, cachedPixmap);
		    }
 
		    if (menuContents == "IconOnly")
		    {
		        insertIconItem(menuPixmap, menuItemId);
                    }
		    else if (menuContents == "TextOnly")
		    {
		        insertTextItem(appName, menuItemId);
		    }
		    else if (menuContents == "TextAndIcon" ||
		             menuContents == "IconAndName")
		    {
		        insertTextAndIconItem(menuPixmap, appName, menuItemId);
		    }
		    else // TaskBar
		    {
		        insertTaskBarItem(menuPixmap, menuItemId);
		    }

		    // Remember that we have created a menu for this exec
		    // Now we use the value later on
		    logMessage("rebuildMenu: remembering channel info");
		    channelMap->insert(channelName, new QString(execName));
 
		    // Remember which id maps to which channel name
		    logMessage("rebuildMenu: itemId=%d", menuItemId);
		    idMap->insert(menuItemId++, new QString(channelName));

	        }
	        // Finished with this config file so cleanup
	        delete cfgFile;
            }
	}
    }

    if (canRaiseDesktop)
    {
        logMessage("rebuildMenu: about to add entry for Qtopia desktop");
	QPixmap *cachedPixmap;
	QPixmap  menuPixmap;
	// Have we already cached the pixmap? (saves us having to 
	// resize the icons every time the menu is popped up)
	if ((cachedPixmap = QPixmapCache::find("tasklist/home")))
	{
	    // Use the cached pixmap
	    menuPixmap = *cachedPixmap;
	}
	else
	{
	    // Otherwise load as a image and smooth scale it to the 
	    // required size.
	    QImage im1 = Resource::loadImage("tasklist/home");
	    im1 = im1.smoothScale(16,16);

	    // Convert the image to a pixmap and cache it
	    menuPixmap.convertFromImage(im1);
	    cachedPixmap = new QPixmap(menuPixmap);
	    QPixmapCache::insert("tasklist/home", cachedPixmap);
	}

	if (menuContents == "IconOnly")
	{
	    insertIconItem(menuPixmap, menuItemId);
	}
	else if (menuContents == "TextOnly")
	{
	    insertTextItem(QString("Desktop"), menuItemId);
	}
	else if (menuContents == "TextAndIcon" ||
		 menuContents == "IconAndName")
	{
	    insertTextAndIconItem(menuPixmap, QString("Desktop"), menuItemId);
	}
	else // TaskBar
	{
	    insertTaskBarItem(menuPixmap, menuItemId);
	}
	logMessage("rebuildMenu: remembering channel info");
	QString desktopChannel("QPE/Application/launcher");
	channelMap->insert(desktopChannel, new QString("launcher"));

	logMessage("rebuildMenu: itemId=%d", menuItemId);
	idMap->insert(menuItemId++, new QString(desktopChannel));
    }

    if (menuContents == "TaskBar")
	taskBarLayout->addStretch(10);

    logMessage("EXIT: rebuildMenu");
}


/***
 *
 * TaskListApplet::insertIconItem()
 *
 */
void TaskListApplet::insertIconItem(QPixmap mPixmap, int mId)
{
    logMessage("ENTER: insertIconItem");
    logMessage("insertIconItem: IconOnly item");
    CustomMenuWidget *cw = new CustomMenuWidget(mPixmap, mId, showCancel);
    logMessage("insertIconItem: Created custom widget");
    menu->insertItem(cw, mId);
    logMessage("insertIconItem: Inserted into menu");
    connect(cw, SIGNAL(raisePicked(int)), this, SLOT(raiseApp(int)));
    connect(cw, SIGNAL(cancelPicked(int)), this, SLOT(quitApp(int)));
    logMessage("insertIconItem: signals connected");
    logMessage("EXIT: insertIconItem");
}


/***
 *
 * TaskListApplet::insertTextItem()
 *
 */
void TaskListApplet::insertTextItem(QString aName, int mId)
{
    logMessage("ENTER: insertTextItem");
    logMessage("insertTextItem: TextOnly item");
    CustomMenuWidget *cw = new CustomMenuWidget(aName, mId, showCancel);
    logMessage("insertTextItem: Created custom widget");
    menu->insertItem(cw, mId);
    logMessage("insertTextItem: Inserted into menu");
    connect(cw, SIGNAL(raisePicked(int)), this, SLOT(raiseApp(int)));
    connect(cw, SIGNAL(cancelPicked(int)), this, SLOT(quitApp(int)));
    logMessage("insertTextItem: signals connected");
    logMessage("EXIT: insertTextItem");
}


/***
 *
 * TaskListApplet::insertTextAndIconItem()
 *
 */
void TaskListApplet::insertTextAndIconItem(QPixmap mPixmap, QString aName, int mId)
{
    logMessage("ENTER: insertTextAndIconItem");
    // If we don't recognise the setting then TextAndIcon
    // is the default catch all
    logMessage("insertTextAndIconItem: TextAndIcon item");
    CustomMenuWidget *cw=new CustomMenuWidget(mPixmap, aName, mId, showCancel);
    logMessage("insertTextAndIconItem: Created custom widget");
    menu->insertItem(cw, mId);
    logMessage("insertTextAndIconItem: Inserted into menu");
    connect(cw, SIGNAL(raisePicked(int)), this, SLOT(raiseApp(int)));
    connect(cw, SIGNAL(cancelPicked(int)), this, SLOT(quitApp(int)));
    logMessage("insertTextAndIconItem: signals connected");
    logMessage("EXIT: insertTextAndIconItem");
}


/***
 *
 * TaskListApplet::insertIconItem()
 *
 */
void TaskListApplet::insertTaskBarItem(QPixmap mPixmap, int mId)
{
    logMessage("ENTER: insertTaskBarItem");
    logMessage("insertTaskBarItem: TaskBar item");
    CustomMenuWidget *cw = new CustomMenuWidget(mPixmap, mId,
                                                false, false, taskBar);
    logMessage("insertTaskBarItem: Created custom widget");
    //taskBarLayout->addWidget(cw);
    logMessage("insertTaskBarItem: Inserted into taskbar");
    connect(cw, SIGNAL(raisePicked(int)), this, SLOT(raiseApp(int)));
    logMessage("insertTaskBarItem: signals connected");
    logMessage("EXIT: insertTaskBarItem");
}


/***
 *
 * TaskListApplet::raiseApp()
 *
 */
void TaskListApplet::raiseApp(int id)
{
    logMessage("ENTER: raiseApp");
    QString *channelName;

    pickedId = id;
    
    if ((channelName = idMap->find(id)) != NULL)
    {
	logMessage("raiseApp: id=%d channel=%s", id, channelName->latin1());

	// Send the message
	if (channel != NULL)
	{
	    QCopChannel::send(channelName->latin1(), "raise()");
	}

    }
    else
    {
        logMessage("raiseApp: couldn't find channel name for id=%d", id);
    }

    // Need to close the popup menu
    if (menuContents == "TaskBar")
        taskBarParent->hide();
    else
	menu->close();

    logMessage("EXIT: raiseApp");
}


/***
 *
 * TaskListApplet::quitApp()
 *
 */
void TaskListApplet::quitApp(int id)
{
    // This routine should not get called from the taskbar version of the menu
    // as there are no cancel buttons displayed
    logMessage("ENTER: quitApp");
    QString *channelName;
    
    if ((channelName = idMap->find(id)) != NULL)
    {
	logMessage("quitApp: id=%d channel=%s", id, channelName->latin1());

	// Now find and terminate the process
	if (useKill)
	{
	    QString *exec = channelMap->find(channelName->latin1());
	    QString execName(exec->latin1());
	    terminateApp(execName);
	}
	else
	{
	    // Send the quit message
	    if (channel != NULL)
	    {
		QCopChannel::send(channelName->latin1(), "quit()");
	    }
	}
    }
    else
    {
        logMessage("raiseApp: couldn't find channel name for id=%d", id);
    }

    // Need to close the popup menu
    if (menuContents == "TaskBar")
        taskBarParent->hide(); // Should never do this with the taskbar view
    else
        menu->close();

    logMessage("EXIT: quitApp");
}


/***
 *
 * TaskListApplet::terminateApp
 *
 */
void TaskListApplet::terminateApp(QString exec)
{
    logMessage("ENTER: terminateApp");

    // If this gets set to true at the end we can try to kill the process
    bool pidFound = false;
    int  pid = -1;

    // Open the directory looking for the subdirs
    QDir procDir = QDir("/proc");
    procDir.setFilter(QDir::Dirs);

    // Get an interator for the list of dirs
    const QFileInfoList *procList = procDir.entryInfoList();
    QFileInfoListIterator procIterator( *procList );
    QFileInfo *procInfo;

    // Now for each directory and until we get the pid we need
    while ( (procInfo = procIterator.current()) && !pidFound)
    {
	// Move the iterator on
	++procIterator;   

        // Only want to look in process id dirs so is this dir a number?
	bool ok = true;
	procInfo->fileName().toInt(&ok);
	if (!ok) continue;

	// Construct the path to the status file
	QString statusPath("/proc/");
	statusPath.append(procInfo->fileName()).append("/status");
	logMessage("terminateApp: about to open %s", statusPath.latin1());
	QFile status(statusPath);
	status.open(IO_ReadOnly);

        // First line should be Name: execname
	QString line;
	status.readLine(line, 80);
	line = line.simplifyWhiteSpace();
	if (line.startsWith("Name:"))
	{
	    logMessage("terminateApp: line=%s", line.latin1());

	    // Remove the "Name: "
	    line = line.remove(0, 6);

	    logMessage("terminateApp: line=%s", line.latin1());

	    // Is this status file the one we want?
	    if (!line.startsWith(exec))
	    {
	        status.close();
	        continue;
	    }
	}
	else
	{
	    // Don't recognise the file so skip on to the next one
	    status.close();
	    continue;
	}

	// Now read down looking for the Pid: line
	while (!pidFound)
	{
	    status.readLine(line, 20);
	    line = line.simplifyWhiteSpace();
	    logMessage("terminateApp: line=%s", line.latin1());
	    if (line.startsWith("Pid:"))
	    {
	        pidFound = true;

		// Strip the "Pid: " off the front
		line = line.remove(0, 4);
	        logMessage("terminateApp: line=%s", line.latin1());

		// Convert to an int
		pid = line.toInt(&ok);
		// Final check....
		if (!ok)
		{
		    // Something went wrong, give up!
		    logMessage("terminateApp: failed to convert to int");
		    pid = -1;
		    pidFound = false;
		    status.close();
		    break;
		}
	    }
	}
	status.close();
    }

    // Do the kill
    if (pidFound && pid != -1)
    {
        logMessage("terminateApp: found pid=%d for exec=%s", pid, exec.latin1());
	logMessage("terminateApp: doing kill(%d, SIGTERM)", pid);
	kill(pid, SIGTERM);
    }
    logMessage("EXIT: terminateApp");
}


/***
 *
 * TaskListApplet::paintEvent()
 *
 */
void TaskListApplet::paintEvent( QPaintEvent* )
{
    logMessage("ENTER: paintEvent");
    printf("TaskListApplet():paintEvent()");

    QPainter p(this);
    p.drawPixmap( 0, (height()-tasklistPixmap.height())/2,
                  tasklistPixmap );
    logMessage("EXIT: paintEvent");
}


/***
 *
 * TaskListApplet::readConfig()
 *
 */
void TaskListApplet::readConfig()
{
    logMessage("ENTER: readConfig");
    // Use a QFile object to check for the existence of the configuration file
    QDir homeDir(QDir::homeDirPath()+"/Settings");
    QFile configFile(homeDir.filePath("TaskListApplet.conf"));

    // And setup a full path to the config file
    QString configFilePath(QDir::homeDirPath()+"/Settings");
    configFilePath.append("/TaskListApplet.conf");
    Config *cfgFile = new Config(configFilePath, Config::File);
    cfgFile->setGroup("Settings");

    if (configFile.exists())
    {
        // Read in our settings
	menuContents      = cfgFile->readEntry("MenuContents");
	showCancel        = cfgFile->readBoolEntry("ShowCancel"); 
	useKill           = cfgFile->readBoolEntry("UseKill", false); 
	taskBarLeftToRight= cfgFile->readBoolEntry("TaskBarLeftToRight", false);
	taskBarTimeout    = cfgFile->readNumEntry("TaskBarTimeout", 5);
	taskBarTimeout   *= 1000;
	debug             = cfgFile->readBoolEntry("Debug", false);
	canRaiseDesktop   = cfgFile->readBoolEntry("CanRaiseDesktop", true);

	// Check debug flag
	if (debug && debugFile == NULL)
	    debugFile = fopen("/tmp/tasklist.log", "a+");
    }
    else
    {
        // Create a config file with our defaults
	cfgFile->writeEntry("MenuContents", menuContents);
	cfgFile->writeEntry("ShowCancel", showCancel);
	cfgFile->writeEntry("UseKill", useKill);
	cfgFile->writeEntry("TaskBarLeftToRight", taskBarLeftToRight);
	cfgFile->writeEntry("TaskBarTimeout", QString::number(taskBarTimeout/1000));
	cfgFile->writeEntry("CanRaiseDesktop", canRaiseDesktop);
	cfgFile->writeEntry("Debug", debug);
    }
    // All done, remove the config file object
    delete cfgFile;

    logMessage("readConfig: MenuContents       = %s", menuContents.latin1());
    logMessage("readConfig: ShowCancel         = %d", showCancel ? 1 : 0);
    logMessage("readConfig: UseKill            = %d", useKill ? 1 : 0);
    logMessage("readConfig: TaskBarLeftToRight = %d", taskBarLeftToRight ? 1 : 0);
    logMessage("readConfig: TaskBarTimeout     = %d", taskBarTimeout/1000);
    logMessage("readConfig: Debug              = %d", debug ? 1 : 0);

    logMessage("EXIT: readConfig");
}


/***
 *
 * TaskListApplet::nextApp()
 *
 */
void TaskListApplet::nextApp()
{
    logMessage("ENTER: nextApp");

    if (pickedId == menu->count()-1)
        pickedId = 0;
    else
    	pickedId++;

    raiseApp(pickedId);

    logMessage("EXIT: nextApp");
}


/***
 *
 * TaskListApplet::nextApp()
 *
 */
void TaskListApplet::prevApp()
{
    logMessage("ENTER: prevApp");

    if (pickedId == 0)
        pickedId = menu->count()-1;
    else
        pickedId--;

    raiseApp(pickedId);

    logMessage("EXIT: prevApp");
}


/***
 *
 * TaskListApplet::hideTaskBar()
 *
 */
void TaskListApplet::hideTaskBar()
{
    if (taskBarParent->isVisible())
       taskBarParent->hide();
}

/***
 *
 * TaskListApplet::signalReceived()
 *
 */
void TaskListApplet::signalReceived(const QCString& msg, const QByteArray& data)
{
    logMessage("ENTER: signalReceived");
    QDataStream stream(data, IO_ReadOnly);
    if (msg == "reload()")
    {
	logMessage("signalReceived: reloading config file");
	readConfig();
    }
    else if (msg == "debug(QString)")
    {
	QString flag;
	stream >> flag;
	flag = flag.lower();
        logMessage("signalReceived: debug(%s)", flag.latin1());

	if (debugFile != NULL)
	{
	    logMessage("signalReceived: closing existing debug file");
	    fclose(debugFile);
	    debugFile = NULL;
	}

	if (flag == "true")
	{
	    debugFile = fopen("/tmp/tasklist.log", "a+");
	    logMessage("signalReceived: debug output enabled");
	}
    }
    else if (msg == "raise()")
    {
        logMessage("signalReceived: raise()");
	raiseApp(pickedId);
    }
    else if (msg == "raiseApp(int)")
    {
        int id;
	stream >> id;
        logMessage("signalReceived: raiseApp(%d)", id);
	raiseApp(id);
    }
    else if (msg == "quitApp(int)")
    {
        int id;
	stream >> id;
        logMessage("signalReceived: quitApp(%d)", id);
	quitApp(id);
    }
    else if (msg == "rebuildMenu()")
    {
        logMessage("signalReceived: rebuildMenu()");
        rebuildMenu();
    }
    else if (msg == "nextApp()")
    {
        logMessage("signalReceived: nextApp()");
	nextApp();
    }
    else if (msg == "prevApp()")
    {
        logMessage("signalReceived: prevApp()");
	prevApp();
    }

    logMessage("EXIT: signalReceived");
}
